// ------------------------------------------------------------------------------
// http://www.arkitus.com
// Copyright (C) 2006  Ali Eslami
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// ------------------------------------------------------------------------------
// PhysicsModel.java
// A more sophisticated version of Model, this agent has built in physics
// attributes and functions.
// ------------------------------------------------------------------------------

class PhysicsModel extends Model
{
  ArrayList forces;
  Vector veloc;
  Vector accel;
  
  float torque;
  float invMomentOfInertia;
  float rotationSpeed;
  float mass;
   
  /**
   * Constructor
   */
  PhysicsModel(String N, String T, Polygon P, Material M, Point C, float R, Settings S, String I)
  {
    super(N, T, P, M, C, R, S, I);
    
    forces = new ArrayList();
    veloc = new Vector();
    accel = new Vector();
    
    torque = 0;
    invMomentOfInertia = calculateMoment();
    rotationSpeed = 0;
    mass = poly.area()/material.invDensity;
  }
   
  /**
   * Executes one frame
   */
  Model frame(ArrayList models, boolean[] keys)
  {
    return this;
  }
  
  /**
   * Moves the object correctly
   */
  void move()
  {
    float delta_t = 1/frameRate;
    
    // ---------------------------
    // Calculate total force on object
    // ---------------------------
    
    Vector totalForceVector = new Vector();
    for (int i=0; i<forces.size(); i++)
    {
      totalForceVector = totalForceVector.add(((Force) forces.get(i)).vector);
    }
    
    // --------------------------
    // Translational movement
    // --------------------------
    
    accel = totalForceVector.multiply(1/mass);
    center = ((veloc.multiply(delta_t)).add(accel.multiply(pow(delta_t,2)/2))).add(center);
    veloc = veloc.add(accel.multiply(delta_t));
    
    // --------------------------
    // Rotational movement
    // --------------------------
    
    for (int i=0; i<forces.size(); i++)
    {
      Force f = ((Force) forces.get(i));
      Vector r = new Vector(f.point);
      torque += f.vector.normal().dot(r);
    }
    
    torque *= invMomentOfInertia;
    float rotationAmount = rotationSpeed*delta_t+torque*(pow(delta_t,2)/2);
    if (abs(rotationAmount) > radians(0.1)) poly = poly.rotate(rotationAmount);
    rotationSpeed += torque*delta_t;
  }
  
  /**
   * Parse the input string
   */
  void parseInput(String input)
  {
  }
  
  /**
   * Calculates the moment of inertia of
   * this model's polygon
   */
  float calculateMoment()
  {
    float R = 0;
    
    if ( material.invDensity != 0 )
    {
      float temp = 0;
      
      for (int i=0; i<poly.numVertex(); i++)
      {
        Vector Pn;
        Vector Pnp1;
        
        Pn = new Vector(poly.vertices[i%poly.numVertex()]);
        Pnp1 = new Vector(poly.vertices[(i+1)%poly.numVertex()]);
        
        temp += (abs(Pn.dot(Pnp1.normal()))*(Pn.dot(Pn)+Pnp1.dot(Pnp1)+Pn.dot(Pnp1)));
      }
      
      R = 1/(temp/material.invDensity/12);
    }
    
    return R;
  }
  
  /**
   * Resets values to get ready for the 
   * next frame
   */
  void reset()
  {
    forces = new ArrayList();
    accel = new Vector();
    torque = 0; 
  }
  
  /**
   * Creates an impulse on the object
   * when it collides with something else
   */  
  void impulse(Vector v, float k, float j, Vector r)
  { 
    float invMass = 1/mass;
    
    center = v.multiply(k).add(center);
    veloc = veloc.add(v.multiply(j*invMass));
    rotationSpeed -= v.multiply(j).dot(r.normal())*invMomentOfInertia;
  }
  
  /**
   * Slows down the movement and rotation
   * artificially
   */
  void damp(float kv, float kr)
  {
    veloc = veloc.multiply(kv);
    rotationSpeed *= kr;
  }
}
