// ------------------------------------------------------------------------------
// http://www.arkitus.com
// Copyright (C) 2006  Ali Eslami
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// ------------------------------------------------------------------------------
// Renderer.java
// An instance of this class will take various inputs from the game engine
// such as the models, the light sources, etc. and display them on the screen.
// ------------------------------------------------------------------------------

class Renderer
{
  Settings settings;
  TrigTable T;
  
  /**
   * Constructor
   */
  Renderer(Settings S)
  {
    settings = S;
    T = new TrigTable(1);
  }
  
  /**
   * Draws the current world onto screen
   */
  void frame(GameWorld gameWorld)
  {
    drawBackground();
    drawModels(gameWorld.models);
  }
  
  /**
   * Draws each model with the correct colours
   */
  void drawModels(ArrayList models)
  {
    // ------------------------------
    // Finding the camera
    // ------------------------------
      
    Model camera = findCamera(models);
    
    for (int i=models.size()-1; i>=0; i--)
    {
      Model model = (Model) models.get(i);
      
      // ------------------------------
      // Layer 1: Particles not on top
      // This section could be redone.
      // ------------------------------

      for (int j=0; j<model.particleManager.particles.size(); j++)
      {
        Particle particle = (Particle) model.particleManager.particles.get(j);
      
        if (!particle.dead && !particle.renderOnTop)
        {
          particle.frame();
          particle.render(settings.PIXELS_PER_METER*(particle.center.x - camera.center.x) + width/2, settings.PIXELS_PER_METER*(particle.center.y - camera.center.y) + height/2);
          model.particleManager.particles.set(j,particle);
        }
      }
        
      // ------------------------------
      // Layer 2: Body Fills
      // ------------------------------
      
      fill(model.material.body);
      noStroke();
      
      beginShape(POLYGON);
      for (int j=0; j<model.poly.numVertex(); j++)
      {
        float x = model.poly.vertices[j].x;
        float y = model.poly.vertices[j].y;
        vertex(settings.PIXELS_PER_METER*(model.center.x + x - camera.center.x) + width/2, settings.PIXELS_PER_METER*(model.center.y + y - camera.center.y) + height/2); 
      }
      endShape();
          
      // ------------------------------
      // Layer 3: Specular Shine
      // ------------------------------
      
      if(model.material.specular != 0)
      {
        for (int j=0; j<models.size(); j++)
        {
          Model glowModel = (Model) models.get(j);
          
          if (j != i && glowModel.material.glow != 0)
            {          
            float distance = pow(model.center.x - glowModel.center.x,2) + pow(model.center.y - glowModel.center.y,2);
            float[] distances = new float[model.poly.numVertex()];
            int kMax = 0;
            float kMaxVal = 0;
          
            // Fill up the distances array from the center of model 
            // to each vertex.
            for (int k=0; k<model.poly.numVertex(); k++)
            {
              distances[k] = pow(glowModel.center.x - model.poly.vertices[k].x - model.center.x,2) + pow(glowModel.center.y - model.poly.vertices[k].y - model.center.y,2);
              
              if ( distances[k] > kMaxVal)
              {
                kMax = k;
                kMaxVal = distances[k];
              }
            }
            
            // Now draw the specular shine
            noStroke();
            fill(255,255,255,((float) glowModel.material.glow/100)*((float) model.material.specular/100)*255);
            
            beginShape(POLYGON);
            for (int k=0; k<distances.length+1; k++)
            {
              float x = model.poly.vertices[(kMax+k)%distances.length].x;
              float y = model.poly.vertices[(kMax+k)%distances.length].y;
              vertex(settings.PIXELS_PER_METER*(model.center.x + x - camera.center.x) + width/2, settings.PIXELS_PER_METER*(model.center.y + y - camera.center.y) + height/2);
            }
            for (int k=distances.length; k>=0; k--)
            {
              if ((kMax+k)%distances.length != kMax)
              {
                float amp = distances[(kMax+k)%distances.length]/distances[kMax];
                if ( amp > 0.95 ) amp = 1;
                if ( amp < settings.MAX_SPECULAR/100 ) amp = settings.MAX_SPECULAR/100;
                float x = model.poly.vertices[(kMax+k)%distances.length].x*amp;
                float y = model.poly.vertices[(kMax+k)%distances.length].y*amp;
                vertex(settings.PIXELS_PER_METER*(model.center.x + x - camera.center.x) + width/2, settings.PIXELS_PER_METER*(model.center.y + y - camera.center.y) + height/2);
              }
            }
            endShape();
          }
        }
      } 
      
      // ------------------------------
      // Layer 4: Borders
      // ------------------------------
      
      if (model.material.stroke != 0)
      {
        noFill();
        stroke(model.material.border);
        strokeWeight(model.material.stroke);
      
        beginShape();
        for (int j=0; j<model.poly.numVertex(); j++)
        {
          float x = model.poly.vertices[j].x;
          float y = model.poly.vertices[j].y;
          vertex(settings.PIXELS_PER_METER*(model.center.x + x - camera.center.x) + width/2, settings.PIXELS_PER_METER*(model.center.y + y - camera.center.y) + height/2); 
        }
        endShape(CLOSE); 
      } 
      
      // ------------------------------
      // Layer 5: Particles on top
      // This section could be redone.
      // ------------------------------

      for (int j=0; j<model.particleManager.particles.size(); j++)
      {
        Particle particle = (Particle) model.particleManager.particles.get(j);
      
        if (!particle.dead && particle.renderOnTop)
        {
          particle.frame();
          particle.render(settings.PIXELS_PER_METER*(particle.center.x - camera.center.x) + width/2, settings.PIXELS_PER_METER*(particle.center.y - camera.center.y) + height/2);
          model.particleManager.particles.set(j,particle);
        }
      }
    }
  }
  
  /**
   * Finds the camera in the list of models
   */
  Model findCamera(ArrayList models)
  {
    for (int i=0; i<models.size(); i++)
    {
      Model model = (Model) models.get(i);
      
      if (model.name.equals("camera"))
      {
        return model;
      }
    }
    
    return null;
  }
  
  /**
   * Draws everything related to the background
   */
  void drawBackground()
  {
    background(255,255,255);
  }
}
